import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;
import javax.imageio.ImageIO;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;

public class DragGlassPane extends JComponent {
  private static final long serialVersionUID = 1L;

  private static BufferedImage image;
  private static BufferedImage shadow;
  private static Point off;

  private Point pos;

  private static List<WeakReference<DragGlassPane>> dgpList =
    new ArrayList<WeakReference<DragGlassPane>>();

  private static ReferenceQueue<DragGlassPane> refQueue =
    new ReferenceQueue<DragGlassPane>();

  private static void cleanupQueue() {
    WeakReference<DragGlassPane> ref;
    while ((ref = (WeakReference<DragGlassPane>) refQueue.poll()) != null) {
      dgpList.remove(ref);
    }
  }

  public DragGlassPane() {
    dgpList.add(new WeakReference<DragGlassPane>(this, refQueue));
  }

  public static void setImage(BufferedImage image, Point off) {
    DragGlassPane.image = image;
    DragGlassPane.off = off;
  }

  public static void showAll() {
    cleanupQueue();

    for (WeakReference<DragGlassPane> ref : dgpList) {
      final DragGlassPane dgp = ref.get();
      if (dgp == null) continue;
      dgp.setVisible(true);
    }
  }
 
  public static void hideAll() {
    image = null;
    off = null;

    cleanupQueue();

    for (WeakReference<DragGlassPane> ref : dgpList) {
      final DragGlassPane dgp = ref.get();
      if (dgp == null) continue;
      dgp.setVisible(false);
      dgp.setInsertionPoint(null);
    }
  }
 
  public static DragGlassPane forDock(Dock dock) {
    if (dock == null) return null;
    final Window w = SwingUtilities.getWindowAncestor(dock.getComponent());
    if (w != null && w instanceof JFrame) {
      return (DragGlassPane) ((JFrame) w).getGlassPane();
    }
    return null;
  }

  public static void updatePositions(Point p) {
    for (WeakReference<DragGlassPane> ref : dgpList) {
      final DragGlassPane dgp = ref.get();
      if (dgp == null) continue;

      final Rectangle bounds = dgp.getBounds();
      bounds.setLocation(dgp.getLocationOnScreen());
      bounds.grow(image.getWidth(), image.getHeight());
      if (bounds.contains(p)) dgp.updatePosition(p);
    } 
  }

  public void updatePosition(Point p) {
    final Point oldPos = pos;
   
    pos = new Point(p); 
    SwingUtilities.convertPointFromScreen(pos, this);

    final int iw = image.getWidth();
    final int ih = image.getHeight();
    
    repaint(pos.x - off.x, pos.y - off.y, iw, ih);
    if (oldPos != null) {
      repaint(oldPos.x - off.x, oldPos.y - off.y, iw, ih);
    }      
  }

/*
  public void moveIt(Point pos) {
    final Point oldPos = this.pos;
    SwingUtilities.convertPointFromScreen(pos, this);
    this.pos = pos;
      
    final int iw = image.getWidth();
    final int ih = image.getHeight();
//    final int sw = shadow.getWidth();
//    final int sh = shadow.getHeight(); 

    final Rectangle clip = new Rectangle(pos.x - off.x, pos.y - off.x, iw, ih);
    if (oldPos != null) {
      clip.add(new Rectangle(oldPos.x - off.x, oldPos.y - off.y, iw, ih));
    }
//    clip.add(new Rectangle(oldPos.x - off.x, oldPos.y - off.y, iw, ih));
//    clip.add(new Rectangle(oldPos.x - iw/2, oldPos.y - ih/2, sw, sh));
//    clip.add(new Rectangle(pos.x - iw/2, pos.y - ih/2, sw, sh));

    repaint(clip);
  }
*/

  public static DragGlassPane getDragGlassPaneAt(Point p) {
    for (WeakReference<DragGlassPane> ref : dgpList) {
      final DragGlassPane dgp = ref.get();
      if (dgp == null) continue;

// FIXME: doesn't deal with window Z-order properly
      final Rectangle bounds = dgp.getBounds();
      bounds.setLocation(dgp.getLocationOnScreen());
      if (bounds.contains(p)) return dgp;
    }
    return null;
  }

  public void setPosition(Point p) {
    SwingUtilities.convertPointFromScreen(p, this);
    pos = p;
  }

/*
  public void showIt(BufferedImage image, Point pos, Point off) {
    this.image = image;
//    shadow = new ShadowRenderer(5, 0.3f, Color.BLACK).createShadow(image);

    SwingUtilities.convertPointFromScreen(pos, this);
    this.pos = pos;
    this.off = off;
      
    setVisible(true);
  }
*/

  private static final BufferedImage arrow; 
  
  static {
    BufferedImage img = null;
    try {
      img = ImageIO.read(DragGlassPane.class.getResource("arrow.png"));
    }
    catch (IOException e) {
      e.printStackTrace();       
    }

    arrow = img;
  }

  private Point insertionPoint = null;

  public void setInsertionPoint(Point p) {
    final Point old_ip = insertionPoint;

    insertionPoint = p;
    if (insertionPoint != null) {
      SwingUtilities.convertPointFromScreen(insertionPoint, this);

      final Point ip = insertionPoint;
      repaint(ip.x-arrow.getWidth()/2, ip.y-arrow.getHeight()+2,
              arrow.getWidth()+1, arrow.getHeight()+1);
    }

    if (old_ip != null) {
      repaint(old_ip.x-arrow.getWidth()/2, old_ip.y-arrow.getHeight()+2,
              arrow.getWidth()+1, arrow.getHeight()+1);
    }
  }

  @Override
  protected void paintComponent(Graphics g) {
    if (image != null && pos != null) {
      if (insertionPoint != null) {
        g.drawImage(arrow, insertionPoint.x-arrow.getWidth()/2,
                           insertionPoint.y-arrow.getHeight()+2, null);
      }

      final int x = pos.x - off.x;
      final int y = pos.y - off.y;
            
//      g.drawImage(shadow, x, y, null);
      g.drawImage(image, x, y, null);
    }
  }
}
