import java.awt.Component;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JSplitPane;

public class SplitDock extends JPanel implements Dock, Dockable {
  private static final long serialVersionUID = 1L;

  // child locations
  public static final String LEFT = "left"; 
  public static final String RIGHT = "right"; 
  public static final String TOP = "top"; 
  public static final String BOTTOM = "bottom"; 

  private Dock parent = null;

  private Dockable first = null;
  private Dockable second = null;

  private JSplitPane split = null;

  public SplitDock() {
    setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
  }

  public Dock getParentDock() {
    return parent;
  }

  public void setParentDock(Dock parent) {
    this.parent = parent;
  }

  public int getDockableCount() {
    return split != null ? 2 : (first != null ? 1 : 0);
  }

  public Dockable[] getDockables() {
    if (split != null) return new Dockable[] { first, second };
    if (first != null) return new Dockable[] { first };
    return new Dockable[0];
  }

  public Component getComponent() {
    return this;
  }

  public void addDockable(Dockable dockable, Object constraint) {
    if (split != null) {
      throw new IllegalArgumentException();
    }
    else if (first != null) {   // we have one child already
      remove(first.getComponent());

      final int orientation =
        LEFT.equals(constraint) || RIGHT.equals(constraint) ?
          JSplitPane.HORIZONTAL_SPLIT : JSplitPane.VERTICAL_SPLIT;

      if (TOP.equals(constraint) || LEFT.equals(constraint)) {
        second = first;
        first = dockable;
      }
      else if (BOTTOM.equals(constraint) || RIGHT.equals(constraint)) {
        second = dockable;
      }
      else {
        throw new IllegalStateException();
      }

      split = new JSplitPane(orientation, true,
                             first.getComponent(), second.getComponent());
      split.setResizeWeight(0.5);
      split.setDividerLocation(0.5);

      add(split);
    }
    else {  // we have no children
      first = dockable;
      add(dockable.getComponent());
    }

    dockable.setParentDock(this);
  }

  public void removeDockable(Dockable child) {
    if (split != null) {  // we have a split
      // replace the split with the remaining child
      final Dockable keeper;
      if      (child == first)  keeper = second;
      else if (child == second) keeper = first;
      else throw new IllegalArgumentException();

      remove(split);
      add(keeper.getComponent());

      split = null;
      first = keeper;
      second = null;
    }
    else if (child == first) {  // this is our sole child
      remove(first.getComponent());
      first = null;
    }
    else {
      throw new IllegalArgumentException();
    }

    child.setParentDock(null);
  }

  public void replaceDockable(Dockable target, Dockable dockable) {
    if (split != null) {
      if (target == first) {
        first = dockable;
        split.setLeftComponent(first.getComponent());
      }
      else if (target == second) {
        second = dockable;
        split.setRightComponent(second.getComponent());
      }
      else throw new IllegalArgumentException();

      target.setParentDock(null);
      dockable.setParentDock(this);
    }
    else if (target == first) {
      removeDockable(target);
      addDockable(dockable, null);
    }
    else {
      throw new IllegalArgumentException();
    }
  }
}
