import java.awt.BasicStroke;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import javax.swing.BoxLayout;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.plaf.TabbedPaneUI;
import javax.swing.plaf.basic.BasicTabbedPaneUI;

public class TabDock extends JTabbedPane implements Dock, Dockable {
  private static final long serialVersionUID = 1L;

  private final DropHandler dh = new DropHandler();
  private final DropTarget dt =
    new DropTarget(this, DnDConstants.ACTION_MOVE, dh, true);

  private final ArrayList<Dockable> dlist = new ArrayList<Dockable>();

  public TabDock() {
  }
 
  private Dock parent = null;
  
  public Dock getParentDock() {
    return parent;
  }

  public void setParentDock(Dock parent) {
    this.parent = parent;
  }
 
  public int getDockableCount() {
    return getTabCount();
  }

  public Dockable[] getDockables() {
    return dlist.toArray(new Dockable[dlist.size()]);
  }

  public Component getComponent() {
    return this;
  }

  public void addDockable(Dockable dockable, Object constraint) {
    addDockable(dockable);
  }

  public void removeDockable(Dockable dockable) {
    remove(dockable.getComponent());
    dlist.remove(dockable);
    dockable.setParentDock(null);
  }

  public void replaceDockable(Dockable target, Dockable dockable) {
    final int index = dlist.indexOf(target);
    removeDockable(target);
    insertDockable(dockable, index);
  }

  public void addDockable(Dockable dockable) {
    insertDockable(dockable, getTabCount());
  }

  public void insertDockable(Dockable dockable, int pos) {
    insertTab("foo", null, dockable.getComponent(), null, pos);
    setTabComponentAt(pos, new DockingHandle("foo", dockable));
/*
    insertTab(dockable.handle.getTitle(), null,
              dockable.getComponent(), null, pos);
    setTabComponentAt(pos, dockable.handle);
*/
    dlist.add(pos, dockable);
    dockable.setParentDock(this);
  }

  private int insertionPoint = -1;

  private Rectangle getTabAreaBounds() {
    final Rectangle r = getUI().getTabBounds(this, getTabCount()-1);
    r.setRect(0, 0, getWidth(), r.y + r.height);
    return r; 
  }

  private int getInsertionIndex(Point pos) {
    if (getTabAreaBounds().contains(pos)) {
      final int count = getTabCount();

      for (int i = 0; i < count; ++i) {
        final Rectangle r = getBoundsAt(i);
        r.x -= r.width/2;
        if (r.contains(pos)) return i;
      }

      return count;
    }
    
    return -1;
  }

  private Point getInsertionLocation(int ip) {
    final Rectangle r;
    if (ip == getTabCount()) {
      r = getBoundsAt(ip-1);
      r.x += r.width;
    }
    else {
      r = getBoundsAt(ip);
    }

    r.y = getTabAreaBounds().y;

    return r.getLocation();
  }
 
  // drop regions
  private static final int NONE   = 0;
  private static final int ABOVE  = 1;
  private static final int RIGHT  = 2;
  private static final int BELOW  = 3;
  private static final int LEFT   = 4;
  private static final int CENTER = 5;
  private static final int TABBAR = 6;
    
  private int dropRegion = NONE;

  private int getRegion(Point p) {
    /*
           ____________
          |\ :     : /|
          |.\:_____:/.|
          |  |     |  |
          |..|_____|..|
          | /:     :\ |
          |/_:_____:_\|
 
    */

    if (getTabAreaBounds().contains(p)) return TABBAR;

    final float w = getWidth();
    final float h = getHeight();

    final float x = p.x;
    final float y = p.y;

// FIXME: Ugly. Clean this up.

    if (x < w/4.0) {                    // left column
      if (y < h/4.0) {                  // top row
        if (x != 0.0 && y/x < h/w) {          // upper wedge
          return ABOVE;
        }
        else {                            // lower wedge
          return LEFT;
        }
      }
      else if (y < 3*h/4) {         // center row
        return LEFT;
      }
      else {                              // bottom row
        if (w/4-x != 0.0 && (y-3*h/4)/(w/4-x) < h/w) {  // upper wedge
          return LEFT;
        }
        else {                            // lower wedge
          return BELOW;
        }
      }
    }
    else if (x < 3*w/4) {           // center column
      if (y < h/4) {                  // top row
        return ABOVE;
      }
      else if (y < 3*h/4) {         // center row
        return CENTER;
      }
      else {                              // bottom row
        return BELOW;
      }
    }
    else {                                // right column
      if (y < h/4) {                  // top row
        if (w-x != 0.0 && y/(w-x) < h/w) {  // upper wedge
          return ABOVE;
        }
        else {                            // lower wedge
          return RIGHT;
        }
      }
      else if (y < 3*h/4) {         // center row
        return RIGHT;
      }
      else {                              // bottom row
        if (x-3*w/4 != 0.0 && (y-3*h/4)/(x-3*w/4) < h/w) {          // upper wedge
          return RIGHT;
        }
        else {                            // lower wedge
          return BELOW;
        }
      }
    }
  }

  private class DropHandler implements DropTargetListener {
    public void dragEnter(DropTargetDragEvent e) {
      final int dr;
      if (getTabCount() == 1 &&
          getDockable(e).getParentDock() == TabDock.this) {
        dropRegion = CENTER;
      }
      else {
        dropRegion = getRegion(e.getLocation());
      }

      damage(dropRegion);

      insertionPoint = getInsertionIndex(e.getLocation());
      if (insertionPoint != -1) {
        final Point p = getInsertionLocation(insertionPoint);
        SwingUtilities.convertPointToScreen(p, TabDock.this);
        DragGlassPane.forDock(TabDock.this).setInsertionPoint(p);
      }  
    }

    public void dragExit(DropTargetEvent e) {
      damage(dropRegion);
      dropRegion = NONE;

      DragGlassPane.forDock(TabDock.this).setInsertionPoint(null);
      insertionPoint = -1;
    }

    public void dragOver(DropTargetDragEvent e) {
      final int dr;
      if (getTabCount() == 1 &&
          getDockable(e).getParentDock() == TabDock.this) {
        dr = CENTER;
      }
      else {
        dr = getRegion(e.getLocation());
      }
      
      if (dr != dropRegion) {
        damage(dropRegion);
        dropRegion = dr;
        damage(dropRegion);
      }

      final int ip = getInsertionIndex(e.getLocation());
      if (insertionPoint != ip) {
        insertionPoint = ip;

        final Point p;  
        if (insertionPoint != -1) {
          p = getInsertionLocation(insertionPoint);
          SwingUtilities.convertPointToScreen(p, TabDock.this);
        }
        else {
          p = null;
        }

        DragGlassPane.forDock(TabDock.this).setInsertionPoint(p);
      }
    }

    public void drop(DropTargetDropEvent e) {
      if (e.isLocalTransfer() &&
          e.isDataFlavorSupported(TransferableDockable.FLAVOR)) {

        final Dockable d = getDockable(e);

        if (dropRegion == NONE) {
          // should be impossible 
          throw new IllegalStateException();
        }

        e.acceptDrop(DnDConstants.ACTION_MOVE);

        final Dock src = d.getParentDock();

        if (dropRegion == CENTER) {
          src.removeDockable(d);
          addDockable(d);
          setSelectedComponent(d.getComponent());
        }
        else if (dropRegion == TABBAR) {
          if (src == TabDock.this) {
            final TabDock tsrc = (TabDock) src;

            final int oi = tsrc.indexOfComponent(d.getComponent());
            if (oi == insertionPoint || oi + 1 == insertionPoint) {
              // do nothing            
            }
            else {
              tsrc.removeDockable(d);
              if (insertionPoint > oi) --insertionPoint;
              insertDockable(d, insertionPoint);
            }
          }
          else {
            src.removeDockable(d);
            insertDockable(d, insertionPoint);
          }

          setSelectedComponent(d.getComponent());
        }
        else {
          // split this TabDock according to the dropRegion
          src.removeDockable(d);

          final TabDock dp = new TabDock();
          dp.addDockable(d);
          dp.setSelectedComponent(d.getComponent());

          final Object constraint;
          switch (dropRegion) {
          case ABOVE: constraint = SplitDock.TOP;    break;
          case RIGHT: constraint = SplitDock.RIGHT;  break;
          case BELOW: constraint = SplitDock.BOTTOM; break;
          case LEFT:  constraint = SplitDock.LEFT;   break;
          default:    throw new IllegalStateException();
          }
          
          if (parent instanceof RootDock || 
              (parent instanceof SplitDock && parent.getDockableCount() == 2)) {
            final SplitDock newSplit = new SplitDock();
            parent.replaceDockable(TabDock.this, newSplit);
            newSplit.addDockable(TabDock.this, null);
            newSplit.addDockable(dp, constraint);
          }
          else {
            ((SplitDock) parent).addDockable(dp, constraint);
          }

          ((Container) parent.getComponent()).invalidate();
        }

        final Dock thisRoot = DockUtils.findRoot(TabDock.this);
        if (src != TabDock.this) {
          final Dock root = DockUtils.findRoot(src);
          DockUtils.collapseTabDocks(root);
          DockUtils.collapseSplitDocks(root);
          if (thisRoot != root) {
            DockUtils.collapseSplitDocks(thisRoot);
            ((Container) root.getComponent()).invalidate();
            root.getComponent().repaint();
          }
        }
        else {
          DockUtils.collapseSplitDocks(thisRoot);
        }

        e.dropComplete(true);
      }
      else {
        e.rejectDrop();
      }

      DragGlassPane.forDock(TabDock.this).setInsertionPoint(null);
      insertionPoint = -1;

      dropRegion = NONE;
      repaint();
    }

    public void dropActionChanged(DropTargetDragEvent e) {}

    private Dockable getDockable(DropTargetDragEvent e) {
      return getDockable(e.getTransferable());
    }

    private Dockable getDockable(DropTargetDropEvent e) {
      return getDockable(e.getTransferable());
    }

    private Dockable getDockable(Transferable tr) {
      try {
        final Object o = tr.getTransferData(TransferableDockable.FLAVOR);
        if (o instanceof Dockable) return (Dockable) o;
      }
      catch (IOException e) {
        // should be impossible
        e.printStackTrace();
      }
      catch (UnsupportedFlavorException e) {
        // should be impossible
        e.printStackTrace();
      }
      return null;
    }
  }

  @Override
  public void paint(Graphics g) {
    super.paint(g);

    if (dropRegion == NONE || dropRegion == TABBAR) return;

    final Graphics2D g2d = (Graphics2D) g.create();

    final int s = 2;
    g2d.setStroke(new BasicStroke(2*s));

    final Rectangle tb = getTabAreaBounds();

    final int w = getWidth();
    final int h = getHeight();

    final int top = tb.height + s;
    final int right = w-s;
    final int bottom = h-s;
    final int left = s;
   
    final int hmid = w/2;
    final int vmid = tb.height + (h - tb.height)/2;

    switch (dropRegion) {
    case ABOVE:
      g2d.drawRect(left, top, right-left, vmid-top);
      break;
    case RIGHT:
      g2d.drawRect(hmid, top, right-hmid, bottom-top);
      break;
    case BELOW:
      g2d.drawRect(left, vmid, right-left, bottom-vmid);
      break;
    case LEFT:
      g2d.drawRect(left, top, hmid-left, bottom-top);
      break;
    case CENTER:
      g2d.drawRect(left, top, right-left, bottom-top);
      break;
    default:
      throw new IllegalStateException();
    }
 
    g2d.dispose();
  }

  private void damage(int region) {
    final int s = 2;

    final Rectangle tb = getTabAreaBounds();

    final int w = getWidth();
    final int h = getHeight();

    final int top = tb.height + s;
    final int right = w-s;
    final int bottom = h-s;
    final int left = s;
   
    final int hmid = w/2;
    final int vmid = tb.height + (h - tb.height)/2;

    switch (region) {
    case ABOVE:
      repaint(left-s, top-s, right-left, s); 
      repaint(right+s, top-s, s, vmid-top);
      repaint(left-s, vmid+s, right-left, s);
      repaint(left-s, top-s, s, vmid-top);
      break;
    case RIGHT:
      repaint(hmid-s, top-s, right-hmid, s); 
      repaint(right+s, top-s, s, bottom-top);
      repaint(hmid-s, bottom+s, right-hmid, s);
      repaint(hmid-s, top-s, s, bottom-top);
      break;
    case BELOW:
      repaint(left-s, vmid-s, right-left, s); 
      repaint(right+s, vmid-s, s, bottom-vmid);
      repaint(left-s, bottom+s, right-left, s);
      repaint(left-s, vmid-s, s, bottom-vmid);
      break;
    case LEFT:
      repaint(left-s, top-s, hmid-left, s); 
      repaint(hmid+s, top-s, s, bottom-top);
      repaint(left-s, bottom+s, hmid-left, s);
      repaint(left-s, top-s, s, bottom-top);
      break;
    case CENTER:
      repaint(left-s, top-s, right-left, s); 
      repaint(right+s, top-s, s, bottom-top);
      repaint(left-s, bottom+s, right-left, s);
      repaint(left-s, top-s, s, bottom-top);
      break;
    default:
      break;
    }
  }
}
